///////////////////////////////////////////////////////////////////////////////////////////////
//----------------------------------------Server32.cpp---------------------------------------//
//																							 //
// This is the 32-bit rendition of the Server.cpp console program. As with the original		 //
// Server.cpp program, this app will connect to the client to exchange data, but this time   //
// a file will be sent, not just a simple message.											 //
//																							 //
// Created by: Drew Sikora and Mike Mahoney													 //
// Created on: 10.15.00																		 //
//																							 //
//-------------------------------------------Notes-------------------------------------------//
//																							 //
// One distinct note would be that the term Server here is used as the endpoint for the data //
// being sent. It's not really a server in that it'll collect data and then send it to other //
// clients. It really is a client, but to offer up a distinction between the sending program //
// and the receiving program, we've opted to call them Client and Server, respectively.		 //
//																							 //
///////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////
//------------------------------------------Includes-----------------------------------------//

// MFC is the devil, exorcise it
#define WIN32_LEAN_AND_MEAN

// Windows includes
#include <windows.h>	// the main Windows header
#include <windowsx.h>	// extra Windows functions

// WinSock includes
#include <winsock2.h>	// the WindSock v2 header file
#include <winsockx.h>	// extra WindSock functions

// other includes
#include <colors.h>		// predefined RGB values for COLORREF params
#include <stdio.h>		// needed for using sprintf()
#include <string.h>		// needed for using strcat()
#include <fstream.h>	// needed for file I/O
#include <memory.h>		//
#include <malloc.h>		// used for memory allocation and manipulation

//-------------------------------------------Defines-----------------------------------------//

// window defines
#define WINDOW_WIDTH	640			// in pixels
#define WINDOW_HEIGHT	480			// in pixels
#define WINDOW_NAME		"Server"	// window class name

// WinSock versions
#define WS_VERSION1	0x0101	// WinSock v1.1
#define WS_VERSION2 0x0202	// WinSock v2.0

// error defines
#define NOWINSOCK	0	// was no a WinSock error
#define WINSOCK		1	// was a WinSock error

// IP defines
#define IP_ADDR		INADDR_ANY	// the client IP address (undefined)
#define PORT_NUM	5000		// the port number

// app states
#define END		0	// the app has ended (sockets closed and not sending data)
#define RUN		1	// the app is still running (sockets still open and sending data)
#define RESTART	2	// the user wants to send file again

// end of file states
#define EOFT	0	// the end of the file is reached
#define NOTEOF	1	// the file still has more to go

// packet defines
#define MAX_PACKET_SIZE	500	// the max packet size in kilobytes

// LineText flag defines
#define HOLD		0	// the text does not move
#define ADVANCE		1	// text moves down one line
#define HADVANCE	3	// used to advance text right after HOLD flag is used

// how long till accept() times out (20 seconds)
#define TIME_OUT	20000

//-------------------------------------------Structures--------------------------------------//

// the first-sent struct heralding the info on the coming data
struct HEADER
{
	LPSTR	filename;	// the name of the file being sent
	LPSTR	location;	// where to create the file
	int		size;		// the size of the file (in bytes)
};

// the actual packet being sent
struct PACKET
{
	int		index;		// the packet index number, used for order confirmation
	LPSTR	data;		// the data read from the file
	int		file;		// tells whether the end of the file is reached
};

//-------------------------------------------Instances---------------------------------------//

// program instances
MSG			msg;		// generic message holder
HDC			gdc;		// handle to the graphic device context
HWND		hwnd;		// main window handle (used to ID app in Windows)
RECT		rect;		// rectangle coords struct, used to validate window
HINSTANCE	ginstance;	// global window instance (used to track app in Windows)

// Winsock instances
SOCKET		Listen;			// our listening or local socket
SOCKET		Port;			// the data socket
WSAData		stWSAData;		// holds WinSock data (i.e. version, max socket #, etc.)
sockaddr	client_sock;	// client socket data 
sockaddr_in	Bind;			// address struct for listening socket

// struct instances
HEADER	header;		// the packet header, always sent first
PACKET	packet;		// the packet itself, which holds the actual file data

// other instances
ofstream out;	// handle for opening a file and performing I/O

//---------------------------------------------Styles----------------------------------------//

// class styles
UINT CSTYLE = CS_OWNDC|		// the window has its own DC handle assigned to it
			  CS_DBLCLKS|	// the window can recognize double mouse clicks
			  CS_HREDRAW|	// the window auto-redraws after any horizontal resize
			  CS_VREDRAW;	// same as above, except only for vertical resize

// window styles
DWORD WSTYLE = WS_OVERLAPPEDWINDOW|	// a window with all the trimmings
			   WS_VISIBLE;			// make the window visible

//--------------------------------------------Globals----------------------------------------//

int		app_state;	// the execution state of the app
int		line_spot;	// the line position to draw new text
DWORD	tick_time;	// the time that the file transfer was started (as measured by the system
					// clock, in milliseconds)
int		pack_index;	// the index counter for the packets
int		data_recv;	// total amount of data received (in bytes), note it is not accurate

//-------------------------------------------Prototypes--------------------------------------//

int Main();
int StartUp();
int CleanUp();
void ErrorHandle(LPCTSTR message, int WSError);
void LineText(HWND hwnd, LPSTR buffer, COLORREF color, int flags);
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

///////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////
//--------------------------------------------WinMain()--------------------------------------//
//																							 //
// WinMain is where it all begins and where it all ends. All other functions branch off and  //
// return to this one. When this function terminates, the entire program terminates.		 //
//																							 //
//--------------------------------------------Variables--------------------------------------//
//																							 //
// hinstance		- the current instance that is used by Windows to track our app			 //
// hprevhinstance	- a field that was used back in the days of Win 3.1 but not today		 //
// lpcmdline		- a null-terminating string that supplies any command line parameters	 //
// ncmdshow			- an integer that indicates how the main app window is to be opened 	 //
//																							 //							 //
int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hprevinstance, LPSTR lpcmdline, int ncmdshow)
{
	// create the window class
	WNDCLASSEX	winclass;

	// fill in the class struct
	winclass.cbSize			= sizeof(WNDCLASSEX);				// size of the struct
	winclass.style			= CSTYLE;							// class style flags
	winclass.lpfnWndProc	= WindowProc;						// the callback function
	winclass.cbClsExtra		= 0;								// extra param (unused)
	winclass.cbWndExtra		= 0;								// extra param (unused)
	winclass.hInstance		= hinstance;						// handle to app instance
	winclass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);	// icon in titlebar (default)
	winclass.hCursor		= LoadCursor(NULL, IDC_ARROW);		// window cursor (default)
	winclass.hbrBackground	= GetStockObject(BLACK_BRUSH);		// paint the window black
	winclass.lpszMenuName	= NULL;								// no menu resource loaded
	winclass.lpszClassName	= WINDOW_NAME;						// name of class
	winclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);	// icon in taskbar (default)

	// register the cass with Windows
	if (!RegisterClassEx(&winclass))
		// display an error message and quit
		ErrorHandle("Could not register class!", NOWINSOCK);

	// create the window
	if (!(hwnd = CreateWindowEx(NULL,			// extra creation param (unused)
								WINDOW_NAME,	// class to derive info from
								"Server",		// text in titlebar
								WSTYLE,			// window style flags
								0, 0,			// initial x, y coords
								WINDOW_WIDTH,	// width of window (in pixels)
								WINDOW_HEIGHT,	// height of window (in pixels)
								NULL, NULL,		// no menu or parent window present
								hinstance,		// app instance handle to attach to
								NULL)))			// extra creation param (unused)
		// display an error message and quit
		ErrorHandle("Could not create window!", NOWINSOCK);
								
	// start up WinSock and do other initialization stuff
	StartUp();
	
	///////////////////////////
	// enter main event loop //
	///////////////////////////

	while(TRUE)
	{
		// test for any messages in the queue
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			// test for a quit message
			if (msg.message == WM_QUIT)
				break;	// exit the loop

			// translate any accelerator keys
			TranslateMessage(&msg);

			// send the message off to the callback function
			DispatchMessage(&msg);
		}

		// test whether or not the program has ended
		if (app_state = END)
			break;	// exit the loop

		// send the file (or continue sending the file)
		Main();

	}

	//////////////////////////
	// exit main event loop //
	//////////////////////////

	// return control to Windows
	return(msg.wParam);

} // end WinMain


///////////////////////////////////////////////////////////////////////////////////////////////
//---------------------------------------WindowProc()----------------------------------------//
//																							 //
// This is the program's callback function, or where all the windows messages (they start    //
// with WM_) are processed. Any message sent by windows like WM_TIMER or WM_PAINT to the	 //
// main program are foisted off onto this function to be dealt with accordingly.			 //
//																							 //
//-----------------------------------------Variables-----------------------------------------//
//																							 //
// hwnd		- the handle of the window the messages are from. Only really matters when you 	 //
//			  are dealing with more than one window.										 //
// msg		- the structure that contains the actual message type along with any extra para- //
//			  meters.																		 //
// wparam	- am extra message parameter													 //
// lparam	- an extra message parameter													 //
//																							 //
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)/////////////////
{
	// use a switch statement to process the message types
	switch(msg)
	{
	case WM_PAINT:		// the window needs repainting for some reason
		{
			// get the DC from Windows
			gdc = GetDC(hwnd);

				// if the app is restarting, clear out the window
				if (app_state == RESTART)
				{
					// get the area of the window
					GetClientRect(hwnd, &rect);

					// fill it in with black
					FillRect(gdc, &rect, CreateSolidBrush(BLACK));
				}

			// release the DC back to Windows
			ReleaseDC(hwnd, gdc);

			// validate the window so we stop getting Paint messages
			GetClientRect(hwnd, &rect);	// get the coords of the client area and store in rect
			ValidateRect(hwnd, &rect);	// validate the given rect (entire client area)

			// return success
			return(0);
		} break;
	case WM_KEYDOWN:	// detects a keypress
		{
			// extract the keycode
			int key_code = (int)wparam;

			// sift through key_code options
			switch(key_code)
			{
			case VK_ESCAPE:	
				{
					// verify the user's request
					int Answer = MessageBox(hwnd, "Are you sure you want to quit?", "", 
											MB_OK | MB_ICONEXCLAMATION);

					// if the user says OK...
					if (Answer == IDOK)
						// ... exit the program
						PostMessage(hwnd, WM_DESTROY, 0, 0);
				} break;
			case VK_RETURN:
				{
					// if the program still runs, ignore this command
					if (app_state == RUN)
						break;

					// set the state to RESTART and repaint the window
					app_state = RESTART;
					UpdateWindow(0);

					// call the StartUp() function
					StartUp();
				} break;
			default: break;
			}

			// return succes
			return(0);
		} break;
	case WM_DESTROY:	// the window has been closed, but the app is still running
		{ 
			// if the program is still running, then cleanup
			if (app_state == RUN)
				CleanUp();

			// post a quit message so the app terminates
			PostQuitMessage(0);

			// return success
			return(0);
		} break;
	default: break;
	}

	// any messages we didn't handle we can let Windows deal with
	return(DefWindowProc(hwnd, msg, wparam, lparam));

} // end WindowProc

///////////////////////////////////////////////////////////////////////////////////////////////
//--------------------------------------------StartUp()--------------------------------------//
//																							 //
// This function will load WinSock and set up all the ports and sockets to allow the sending //
// of data to the Server.																	 //
//																							 //
int StartUp()//////////////////////////////////////////////////////////////////////////////////
{
	// set the program state to RUN
	app_state = RUN;
	
	// set the line spot to zero so text begins at the top of the window
	line_spot = 0;

	// reset the data_sent variable to zero
	data_recv = 0;

	// update the user
	LineText(hwnd, "Loading WinSock...", RED, ADVANCE);

	// load the WinSock DLL
	if (WSAStartup(WS_VERSION2, &stWSAData))
		// display an error message and quit
		ErrorHandle("Could not load WinSock!" , WINSOCK);

	// update the user
	LineText(hwnd, "Creating socket...", RED, ADVANCE);

	// create and test the socket
	Listen = socket(AF_INET, SOCK_STREAM, 0);
	if (Listen == INVALID_SOCKET)
		// display an error message and quit
		ErrorHandle("Could not create socket!", WINSOCK);

	// update the user
	LineText(hwnd, "Socket created successfully.", RED, ADVANCE);
	LineText(hwnd, "Attempting to bind socket to port...", RED, ADVANCE);

	// fill out the socket info struct
	Bind.sin_family			= AF_INET;				// address internet family
	Bind.sin_port			= htons(PORT_NUM);		// port number (must match Servers)
	Bind.sin_addr.s_addr	= htonl(INADDR_ANY);	// Server IP address

	// bind the socket to the port so we can listen
	if (bind(Listen, (LPSOCKADDR)&Bind, sizeof(Bind)))
		// display an error message and quit
		ErrorHandle("Could not bind socket!", WINSOCK);

	// listen for the client 
	if (listen(Listen, 1))
		// display an error message and quit
		ErrorHandle("Could not listen on socket!", WINSOCK);

	// update the user
	LineText(hwnd, "Socket has been binded. Listening for client...", RED, ADVANCE);

	// get the accept loop start time
	tick_time = GetTickCount();

	// try to connect for twenty seconds before erring out. Then ask the user if he wants to
	// try again.
	while (TRUE)
	{
		// define variables
		int answer;								// holds answer to message box
		int lenclient = sizeof(client_sock);	// holds size of client struct

		while (((GetTickCount()-tick_time) < TIME_OUT) || (Port != INVALID_SOCKET))
		{
			// connect to the server on Port
			Port = accept(Listen, &client_sock, &lenclient);
		}
			
		// display an error message if result != 1, else break out of the loop
		if (Port == INVALID_SOCKET)
		{
			answer = MessageBox(hwnd, "Could not connect to Client. Connection may have timed out. Try again?",
								"WinSock Error!", MB_YESNO | MB_ICONEXCLAMATION);
		}
		else 
		{
			break;
		}

		// now decide whether or not the user wants to try again or quit
		if (answer == IDNO)
		{
			// cleanup and then exit the function
			CleanUp();
			return(0);
		}
	}

	// update the user
	LineText(hwnd, "Connected to Client! Receiving file header.", RED, ADVANCE);			

	// get the size of the header 
	int header_size = sizeof(HEADER);

	// create a buffer to hold the info
	char *lpBuff = new char[header_size];

	// retrieve the HEADER struct 
	if (recv(Port, lpBuff, header_size, 0) == SOCKET_ERROR)
		// display an error message and quit
		ErrorHandle("Could not receive packet header!", WINSOCK);

	// copy the buffer data into the structure
	//memcpy(header, lpBuff, header_size);
	
	LineText(hwnd, "Creating file for write...", RED, ADVANCE);

	// use the header info to construct the file path and store it in fpath
	LPSTR fpath = header.location;
	strcat(fpath, header.filename);

	// Create and open the file
	out.open(fpath, ios::binary);

	// update the user
	LineText(hwnd, "File has been created successfully.", RED, ADVANCE);

	// tell the user where the file was created and that transfer is starting
	char file[100];
	sprintf(file, "File location: %s", fpath);
	LineText(hwnd, file, RED, ADVANCE);
	LineText(hwnd, "Beginning file transfer...", RED, ADVANCE);

	// get the starting time of the transfer
	tick_time = GetTickCount();

} // end StartUp

///////////////////////////////////////////////////////////////////////////////////////////////
//---------------------------------------------Main()----------------------------------------//
//																							 //
// This function has the duty of reading more and more data from the file and then sending   //
// it off to the Server. This function alone decides when the program state turns to END,	 //
// that happens when the end of the file is reached.										 //
//																							 //
int Main()/////////////////////////////////////////////////////////////////////////////////////
{
	// get the size of PACKET
	int packet_size = sizeof(PACKET);

	// create a buffer to hold the information
	char *lpBuff = new char[packet_size];

	// receive data from the Server
	if (recv(Port, lpBuff, packet_size, 0) == SOCKET_ERROR)
		// display an error message and quit
		ErrorHandle("Could not receive packet!", WINSOCK);

	// copy the data from the buffer to the packet struct
	//memcpy(packet, &lpBuff, packet_size);

	// write to the file
	out.write(packet.data, sizeof(packet.data));

	// increment the amount of data received
	data_recv += MAX_PACKET_SIZE;

	// update the user on what was received
	char recv[50];
	sprintf(recv, "%i bytes received out of %i bytes.", data_recv, header.size);
	LineText(hwnd, recv, RED, HOLD);

	// decide whether the program has ended
	if (packet.file == EOFT)
		CleanUp();

	// return success
	return(0);

} // end Main

///////////////////////////////////////////////////////////////////////////////////////////////
//--------------------------------------------CleanUp()--------------------------------------//
//																							 //
// This function will unload WinSock after it closes all the ports and sockets and dis-		 //
// connects from the server.																 //
//																							 //
int CleanUp()//////////////////////////////////////////////////////////////////////////////////
{
	// get the total alotted time that the transfer took and convert it into seconds
	int time = (GetTickCount() - tick_time);
	
	// close the sockets (listening and client) and tell the user
	closesocket(Port);
	closesocket(Listen);
	LineText(hwnd, "Sockets have been shutdown", RED, HADVANCE);

	// cleanup by unloading the WinSock DLL and informing the user
	WSACleanup();
	LineText(hwnd, "WinSock has been unloaded.", RED, ADVANCE);

	// close the file
	out.close();

	// update the user
	LineText(hwnd, "File has been closed.", RED, ADVANCE);

	// display the time it took to transfer the file
	LPSTR strTime;
	sprintf(strTime, "Time of file transfer: %i", time);
	LineText(hwnd, strTime, RED, ADVANCE);

	// let the user restart the program by pressing Enter
	LineText(hwnd, "To test again, press Enter. Or press Escape to quit.", RED, ADVANCE);

	// set the program state to END
	app_state = END;

	// return success
	return(0);

} // end CleanUp

///////////////////////////////////////////////////////////////////////////////////////////////
//------------------------------------------ErrorHandle()------------------------------------//
//																							 //
// This function will handle an errors that might crop up during program execution			 //
//																							 //
//--------------------------------------------Variables--------------------------------------//
//																							 //
// message	- the message to display in the message box (for non-WinSock errors only)		 //
// WSError	- the error was from WinSock, use WSAGetLastError()								 //
//																							 //
void ErrorHandle(LPCTSTR message, int WSError)/////////////////////////////////////////////////
{
	// define the message box style
	DWORD ESTYLE = MB_OK | MB_ICONEXCLAMATION;

	// decide whether the error was caused by WinSock
	switch(WSError)
	{
	case NOWINSOCK:
		{
			// let the user know what went wrong
			MessageBox(hwnd, message, "Program Error!", ESTYLE);
		} break;
	case WINSOCK:
		{
			// cast the error number
			LPCTSTR err_code = (LPCTSTR)WSAGetLastError();

			// use WSAGetLastError() to display the error code
			MessageBox(hwnd, err_code, "WinSock Error!", ESTYLE);
		} break;
	default: 
		{
			// ???
			MessageBox(hwnd, "An unknown error has occured", "Program Error!", ESTYLE);
		} break;
	}

	// exit the program
	PostMessage(hwnd, WM_DESTROY, 0, 0);

} // end ErrorHandle

///////////////////////////////////////////////////////////////////////////////////////////////
//-------------------------------------------LineText()--------------------------------------//
//																							 //
// This program will space text for the program - every 15 pixels is a new line of text. It  //
// can also hold the text to one line so that a counter can increment in realtime. However,  //
// while the HOLD flag is set, no other text can be written.								 //
//																							 //
//-------------------------------------------Variables---------------------------------------//
//																							 //
// hwnd		- the window handle																 //
// buffer	- the message to output															 //
// color	- text color																	 //
// flags	- behavior flags, hold a line or advance to the next							 //
//																							 //
void LineText(HWND hwnd, LPSTR buffer, COLORREF color, int flags)//////////////////////////////
{
	// define the spacing between lines
	#define LINE 15

	// define variables for text drawing
	int			old_bkmode;
	COLORREF	old_tcolor;

	// get the DC from Windows so we can write text onscreen
	gdc = GetDC(hwnd);

	// set the text color and the background mode (transparent) and store the old 
	// values so we can reset them later on.
	old_tcolor = SetTextColor(gdc, color);
	old_bkmode = SetBkMode(gdc, TRANSPARENT);

	// if the HOLD flag is set, then clear out the text area
	if (flags == HOLD)
	{
		// get the client area in case the window was resized horizontally
		GetClientRect(hwnd, &rect);

		// reset the other values to narrow the rect to just the line of text to erase
		rect.left	= 0;
		rect.top	= line_spot + 3;
		rect.bottom	= line_spot + LINE;

		// erase the text with a solid black rectangle
		FillRect(gdc, &rect, CreateSolidBrush(BLACK));
	}

	// since the if statement at the end is needed so that the HOLD command can work without
	// advancing an extra line, this line is needed because if ADVANCE were used right after a
	// HOLD, the line of text would not advance until after it were printed. 
	if (flags == HADVANCE)
		line_spot += LINE;

	// print out the text
	TextOut(gdc, 0, line_spot, buffer, strlen(buffer));

	// reset the old values
	SetTextColor(gdc, old_tcolor);
	SetBkMode(gdc, old_bkmode);

	// release the DC back to Windows
	ReleaseDC(hwnd, gdc);

	// move down a line if the ADVANCE flag is set and the HADVANCE is not, because if it is,
	// then the line was already advanced.
	if (flags == ADVANCE && flags != HADVANCE)
		line_spot += LINE;

} // end LineText


///////////////////////////////////////////////////////////////////////////////////////////////
//------------------------------------------Server32.cpp-------------------------------------//
//--------------------------------Copyright Napali Networks, 2000----------------------------//
///////////////////////////////////////////////////////////////////////////////////////////////